/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*- */
/*
 * Copyright (C) 2010 ammonkey
 *
 * Nautilus is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Nautilus is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; see the file COPYING.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author: ammonkey <am.monkeyd@gmail.com>
 *
 */

#include <config.h>
#include "nautilus-search-engine-zeitgeist.h"
#include <glib.h>
#include <gmodule.h>
#include <gio/gio.h>
#include <eel/eel-gtk-macros.h>
#include <eel/eel-glib-extensions.h>
#include <gmodule.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <zeitgeist.h>
#include <time.h>

#if 0
typedef struct _ZeitgeistLog ZeitgeistLog;
typedef struct _ZeitgeistTimeRange ZeitgeistTimeRange;
typedef struct _ZeitgeistStorageState ZeitgeistStorageState;
typedef struct _ZeitgeistResultType ZeitgeistResultType;

static void (*zeitgeist_log_find_events)        (ZeitgeistLog        *self,
                                                 ZeitgeistTimeRange  *time_range,
                                                 GPtrArray           *event_templates,
                                                 ZeitgeistStorageState storage_state,
                                                 guint32              num_events,
                                                 ZeitgeistResultType  result_type,
                                                 GCancellable        *cancellable,
                                                 GAsyncReadyCallback  callback,
                                                 gpointer             user_data) = NULL;

static gboolean 
open_libzeitgeist (void)
{
        gpointer x;

        int i;
        GModule *zeitgeist;
        GModuleFlags flags;

        flags = G_MODULE_BIND_LAZY | G_MODULE_BIND_LOCAL;

        zeitgeist = g_module_open ("libzeitgeist-1.0.so.0", flags);

        if (!zeitgeist)
                return FALSE;
        if(g_module_symbol(zeitgeist, "zeitgeist_log_find_events", (gpointer *) &zeitgeist_log_find_events) == FALSE) {
                g_error("Unable to get function reference: %s", g_module_error());
                return FALSE;
        }
#if 0
        for (i = 0; i < G_N_ELEMENTS (tracker_dl_mapping); i++) {
                if ((tracker_dl_mapping[i].versions & version) == 0)
                        continue;

                if (!g_module_symbol (tracker, tracker_dl_mapping[i].fn_name,
                                      tracker_dl_mapping[i].fn_ptr_ref)) {
                        g_warning ("Missing symbol '%s' in libtracker\n",
                                   tracker_dl_mapping[i].fn_name);
                        g_module_close (tracker);

                        for (i = 0; i < G_N_ELEMENTS (tracker_dl_mapping); i++)
                                tracker_dl_mapping[i].fn_ptr_ref = NULL;

                        return 0;
                }
        }
#endif
        return TRUE;
}
#endif

struct NautilusSearchEngineZeitgeistDetails {
        NautilusQuery 	*query;
        gboolean 	query_pending;
        GCancellable    *cancellable;
        ZeitgeistLog    *log;
        GtkWidget       *filterbar;
        GList           *btns;
        char            *search_text;
        GRegex          *regex;
        ZeitgeistResultSet      *events;
        GList           *subj_selection;
};


static void  nautilus_search_engine_zeitgeist_class_init       (NautilusSearchEngineZeitgeistClass *class);
static void  nautilus_search_engine_zeitgeist_init             (NautilusSearchEngineZeitgeist      *engine);
static void  filterbar_reset                                   (NautilusSearchEngineZeitgeist *zeitgeist);

G_DEFINE_TYPE (NautilusSearchEngineZeitgeist,
               nautilus_search_engine_zeitgeist,
               NAUTILUS_TYPE_SEARCH_ENGINE);

static NautilusSearchEngineClass *parent_class = NULL;

static void
finalize (GObject *object)
{
        NautilusSearchEngineZeitgeist *zeitgeist;

        zeitgeist = NAUTILUS_SEARCH_ENGINE_ZEITGEIST (object);
        filterbar_reset (zeitgeist);

        if (zeitgeist->details->query) {
                g_object_unref (zeitgeist->details->query);
                //zeitgeist->details->query = NULL;
        }

        g_object_unref (zeitgeist->details->cancellable);
        //zeitgeist->details->cancellable = NULL;
        //g_object_unref (zeitgeist->details->log);
        //g_free (zeitgeist->details);
        EEL_CALL_PARENT (G_OBJECT_CLASS, finalize, (object));
}

static gboolean
search_thread_done_idle (gpointer user_data)
{
        /*NautilusSearchEngineZeitgeist *zeitgeist =
                NAUTILUS_SEARCH_ENGINE_ZEITGEIST (user_data);*/

        //if (!g_cancellable_is_cancelled (zeitgeist->details->cancellable)) {
                //nautilus_search_engine_finished (NAUTILUS_SEARCH_ENGINE (zeitgeist));
                nautilus_search_engine_finished (NAUTILUS_SEARCH_ENGINE (user_data));
                //data->engine->details->active_search = NULL;
        //}
        return FALSE;
}

/*
static gboolean
search_thread_add_hits_idle (gpointer user_data)
{
        NautilusSearchEngineZeitgeist *zeitgeist = user_data;

        if (!g_cancellable_is_cancelled (zeitgeist->details->cancellable))
                nautilus_search_engine_hits_added (NAUTILUS_SEARCH_ENGINE (zeitgeist), zeitgeist->details->hit_uris);
        return FALSE;
}*/

static gboolean
string_search_filter_events(char *s, GRegex *regex)
{
        //GMatchInfo *match_info;

        //g_regex_match_full (regex, string, -1, 0, 0, &match_info, &error);
        //return (g_regex_match (regex, s, 0, &match_info));
        return (g_regex_match (regex, s, 0, NULL));
}

typedef struct
{
        char    *interpretation;
        gint    count;
} InterpretationFilter;

static GList *
add_interpretation(GList *interpretations, char *s)
{
        InterpretationFilter *ipf;
        GList *l = interpretations;

        for (; l != NULL; l=l->next)
        {
                ipf = l->data;
                if (!strcmp(ipf->interpretation, s))
                {
                        ipf->count++;
                        return (interpretations);
                }
        }
        ipf = g_new0 (InterpretationFilter, 1);
        ipf->interpretation = s;
        ipf->count = 1;
        interpretations = g_list_prepend (interpretations, ipf);
        return (interpretations);
}

static void
button_clicked_callback (GtkWidget *button, NautilusSearchEngineZeitgeist *zeitgeist)
{
        GList *hit_uris = NULL; 
        char *uri;
        char *interpretation;

        GList *l = zeitgeist->details->subj_selection;
        for (; l != NULL; l=l->next)
        {
                uri =  (char *)zeitgeist_subject_get_uri (l->data);
                interpretation = (char *) zeitgeist_subject_get_interpretation (l->data);
                if (!strcmp (interpretation, (char *) g_object_get_data (G_OBJECT (button), "interpretation")))
                        hit_uris = g_list_prepend (hit_uris, uri);
        }

        if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button)))
                nautilus_search_engine_hits_added (NAUTILUS_SEARCH_ENGINE (zeitgeist), hit_uris);
        else
                nautilus_search_engine_hits_subtracted (NAUTILUS_SEARCH_ENGINE (zeitgeist), hit_uris);
        
        g_list_free (hit_uris);
}

static void
filter_bar_button_append (NautilusSearchEngineZeitgeist *zeitgeist, gchar *label, gchar *interpretation)
{
        GtkWidget *btn;
        GtkWidget *hbox = zeitgeist->details->filterbar;
        
        btn = gtk_toggle_button_new_with_label(label);
        gtk_button_set_relief (GTK_BUTTON (btn), GTK_RELIEF_NONE);
        gtk_box_pack_start (GTK_BOX (hbox), btn, FALSE, FALSE, 0);
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (btn), TRUE);
        gtk_widget_show (btn);
        zeitgeist->details->btns = g_list_prepend(zeitgeist->details->btns, btn);
       
        //g_object_set_data (G_OBJECT (btn), "interpretation", strdup (interpretation));
        g_object_set_data (G_OBJECT (btn), "interpretation", interpretation);
        g_signal_connect (btn,
                          "toggled",
			  G_CALLBACK (button_clicked_callback),
			  zeitgeist);
}

static void
filterbar_reset (NautilusSearchEngineZeitgeist *zeitgeist)
{
        GList *l = zeitgeist->details->btns;

        for (; l!= NULL; l=l->next)
                gtk_widget_destroy(l->data);
        zeitgeist->details->btns = NULL;
}

static void
filterbar_set_interpretations(NautilusSearchEngineZeitgeist *zeitgeist, GList *interpretations)
{
        InterpretationFilter *ipf;
        gchar **res;
        gchar *label;

        filterbar_reset (zeitgeist);
        GRegex *regex = g_regex_new ("#", 0, 0, NULL);
        g_regex_ref (regex);
        GList *l = interpretations;
        for (; l != NULL; l=l->next)
        {
                ipf = l->data;
                //printf("%s: %d\n", ipf->interpretation, ipf->count);
                res = g_regex_split (regex, ipf->interpretation, 0);
                //printf("%s: %d\n", res[1], ipf->count);
                label = malloc (strlen(res[1]) + 9);
                sprintf (label, "%s (%d)", res[1], ipf->count);
                //printf ("%s %s\n", label, ipf->interpretation);
                filter_bar_button_append (zeitgeist, label, ipf->interpretation);
                //printf ("%s\n", label);
                g_free (label);
        }
        g_strfreev(res);
        g_regex_unref (regex);
}

static void
on_events_received (ZeitgeistLog  *log,
                    GAsyncResult *res,
                    gpointer      user_data)
{
        ZeitgeistResultSet      *events;
        ZeitgeistEvent          *event;
        ZeitgeistSubject        *subject;
        GError                  *error;
        gint                    i, j;
        char                    *uri;
        GString                 *l_uri;
        char                    *subject_interpretation;

        error = NULL;
        GList *hit_uris = NULL;
        GList *interpretations = NULL;

        NautilusSearchEngineZeitgeist *zeitgeist = NAUTILUS_SEARCH_ENGINE_ZEITGEIST (user_data);
        
        events = zeitgeist_log_find_events_finish (log, res, &error);
        if (zeitgeist->details->events != NULL)
                g_object_unref (zeitgeist->details->events);
        zeitgeist->details->events = events;
        if (zeitgeist->details->subj_selection)
                g_list_free (zeitgeist->details->subj_selection);
        zeitgeist->details->subj_selection = NULL;

        if (error)
        {
                g_warning ("Error reading results: %s", error->message);
                g_error_free (error);
                return;
        }

        //g_message ("Got %i events:", zeitgeist_result_set_size (events));
        //g_message ("search_text: %s", zeitgeist->details->search_text);

        while (zeitgeist_result_set_has_next (events))
        {
                event = zeitgeist_result_set_next (events);
                for (j = 0; j < zeitgeist_event_num_subjects (event); j++)
                {
                        subject = zeitgeist_event_get_subject (event, j);
                        uri =  (char *)zeitgeist_subject_get_uri (subject);
                        GString *stru = g_string_new (uri);
                        l_uri = g_string_ascii_down (stru);
                        /* we got an empty string search let's show all the events */
                        if (strlen (zeitgeist->details->search_text) < 1)
                        {
                                hit_uris = g_list_prepend (hit_uris, uri);
                                subject_interpretation = (char *) zeitgeist_subject_get_interpretation (subject);
                                if (subject_interpretation != NULL)
                                {
                                        interpretations = add_interpretation (interpretations, subject_interpretation);
                                        zeitgeist->details->subj_selection = g_list_prepend (zeitgeist->details->subj_selection, subject);
                                }
                        }
                        else
                                if (string_search_filter_events(l_uri->str, zeitgeist->details->regex))
                                {
                                        hit_uris = g_list_prepend (hit_uris, uri);
                                        subject_interpretation = (char *) zeitgeist_subject_get_interpretation (subject);
                                        if (subject_interpretation != NULL)
                                        {
                                                interpretations = add_interpretation (interpretations, subject_interpretation);
                                                zeitgeist->details->subj_selection = g_list_prepend (zeitgeist->details->subj_selection, subject);
                                        }
                                }
                        //g_string_free (stru, TRUE);
                        g_string_free (l_uri, TRUE);
                }
        }
       
        filterbar_set_interpretations(zeitgeist, interpretations);
        nautilus_search_engine_hits_added (NAUTILUS_SEARCH_ENGINE (zeitgeist), hit_uris);

        g_list_free (hit_uris);
}

#if 0
static void 
get_today(time_t *t_start, time_t *t_end)
{
        *t_end = time(0);
        struct tm *tm = localtime(t_end);
        tm->tm_sec = 0;
        tm->tm_min = 0;
        tm->tm_hour = 0;
        
        //tm->tm_mday = tm->tm_mday -1;
        /*char *ct = asctime(tm);
        printf("t c %s\n", ct);*/
        *t_start = mktime(tm);
}
#endif

static gpointer 
search_thread_func (gpointer user_data)
{
        NautilusSearchEngineZeitgeist *zeitgeist = NAUTILUS_SEARCH_ENGINE_ZEITGEIST (user_data);

        GPtrArray       *templates;
        gint64          tstart, tend;

        //zeitgeist->details->query_pending = FALSE;
        gint period = nautilus_query_get_period (zeitgeist->details->query);
        zeitgeist->details->filterbar = nautilus_query_get_filterbar (zeitgeist->details->query);
        
        /* convert uri and search string to lower char */
        zeitgeist->details->search_text = nautilus_query_get_text (zeitgeist->details->query);
        GString* l_search_text = g_string_ascii_down (g_string_new(zeitgeist->details->search_text));
        zeitgeist->details->regex = g_regex_new (l_search_text->str, 0, 0, NULL);
        g_regex_ref (zeitgeist->details->regex);

        //printf("period: %d\n", period);
        templates = g_ptr_array_new ();
        g_ptr_array_ref (templates);
        g_ptr_array_add (templates, zeitgeist_event_new ());
        
        tend = zeitgeist_timestamp_for_now (); 
        tstart = zeitgeist_timestamp_prev_midnight (tend) -
                (period-1)*ZEITGEIST_TIMESTAMP_DAY;
        ZeitgeistTimeRange *time_range = zeitgeist_time_range_new (tstart, tend);

        zeitgeist_log_find_events (zeitgeist->details->log,
                                   time_range,
                                   templates,
                                   ZEITGEIST_STORAGE_STATE_ANY,
                                   -1,
                                   //ZEITGEIST_RESULT_TYPE_MOST_RECENT_EVENTS,
                                   ZEITGEIST_RESULT_TYPE_MOST_RECENT_SUBJECTS,
                                   //ZEITGEIST_RESULT_TYPE_MOST_POPULAR_SUBJECTS,
                                   //ZEITGEIST_RESULT_TYPE_LEAST_POPULAR_SUBJECTS,
                                   zeitgeist->details->cancellable,
                                   (GAsyncReadyCallback)on_events_received,
                                   zeitgeist);

        /*char *uri = g_strdup("file:///home/am/Musique/VA-Kick-Ass_OST-2010-FRAY/14-elvis_presley-an_american_trilogy.mp333");
        hit_uris = g_list_prepend (hit_uris, (char *)uri);
        zeitgeist->details->hit_uris = hit_uris;*/

        /*if (!g_cancellable_is_cancelled (zeitgeist->details->cancellable)) {
          nautilus_search_engine_hits_added (NAUTILUS_SEARCH_ENGINE (zeitgeist), hit_uris);
          nautilus_search_engine_finished (NAUTILUS_SEARCH_ENGINE (zeitgeist));
        }*/

        //g_idle_add (search_thread_add_hits_idle, zeitgeist);
        g_idle_add (search_thread_done_idle, zeitgeist);
        zeitgeist->details->query_pending = FALSE;

        g_ptr_array_unref (templates);
        //g_ptr_array_free (templates, TRUE);
        
        g_free (zeitgeist->details->search_text);
        g_string_free (l_search_text, TRUE);
        g_regex_unref (zeitgeist->details->regex);
        
        return NULL;
}

static void
nautilus_search_engine_zeitgeist_start (NautilusSearchEngine *engine)
{
        NautilusSearchEngineZeitgeist *zeitgeist;
        GList 	*mimetypes, *l;
        //char    *location, *location_uri;
        //char 	**mimes;
        //int 	i, mime_count;

        //printf("%s\n", G_STRFUNC);
        zeitgeist = NAUTILUS_SEARCH_ENGINE_ZEITGEIST (engine);


        if (zeitgeist->details->query_pending) {
                return;
        }

        if (zeitgeist->details->query == NULL) {
                return;
        }

        g_cancellable_reset (zeitgeist->details->cancellable);
        zeitgeist->details->log = zeitgeist_log_new();
        g_object_ref (zeitgeist->details->log);

        /*mimetypes = nautilus_query_get_mime_types (zeitgeist->details->query);

          location_uri = nautilus_query_get_location (zeitgeist->details->query);*/

        /* we don't need location for a zeitgeist search 
        ** TODO remove the option in the extra bar */
        /*if (location_uri) {
          location = g_strdup (location_uri);
        //g_filename_from_uri (location_uri, NULL, NULL) :
        g_free (location_uri);
        } else {
        location = NULL;
        }*/

        //mime_count = g_list_length (mimetypes);

        g_thread_create (search_thread_func, zeitgeist, FALSE, NULL);

        //g_free (location);
        zeitgeist->details->query_pending = TRUE;
        //eel_g_list_free_deep (mimetypes);
}

static void
nautilus_search_engine_zeitgeist_stop (NautilusSearchEngine *engine)
{
        NautilusSearchEngineZeitgeist *zeitgeist;

        zeitgeist = NAUTILUS_SEARCH_ENGINE_ZEITGEIST (engine);

        //printf(">>%s\n", G_STRFUNC);
        if (zeitgeist->details->query && zeitgeist->details->query_pending) {
                //zeitgeist_cancel_last_call (zeitgeist->details->client);
                //g_cancellable_cancel (zeitgeist->details->cancellable);
                zeitgeist->details->query_pending = FALSE;
        }
        g_cancellable_cancel (zeitgeist->details->cancellable);
        g_object_unref (zeitgeist->details->log);
}

static gboolean
nautilus_search_engine_zeitgeist_is_indexed (NautilusSearchEngine *engine)
{
        return TRUE;
}

static void
nautilus_search_engine_zeitgeist_set_query (NautilusSearchEngine *engine, NautilusQuery *query)
{
        //printf ("::%s\n", G_STRFUNC);
        NautilusSearchEngineZeitgeist *zeitgeist;

        zeitgeist = NAUTILUS_SEARCH_ENGINE_ZEITGEIST (engine);

        if (query) {
                g_object_ref (query);
        }

        if (zeitgeist->details->query) {
                g_object_unref (zeitgeist->details->query);
        }

        zeitgeist->details->query = query;
}

static void
nautilus_search_engine_zeitgeist_class_init (NautilusSearchEngineZeitgeistClass *class)
{
        GObjectClass *gobject_class;
        NautilusSearchEngineClass *engine_class;

        parent_class = g_type_class_peek_parent (class);

        gobject_class = G_OBJECT_CLASS (class);
        gobject_class->finalize = finalize;

        engine_class = NAUTILUS_SEARCH_ENGINE_CLASS (class);
        engine_class->set_query = nautilus_search_engine_zeitgeist_set_query;
        engine_class->start = nautilus_search_engine_zeitgeist_start;
        engine_class->stop = nautilus_search_engine_zeitgeist_stop;
        engine_class->is_indexed = nautilus_search_engine_zeitgeist_is_indexed;
}

static void
nautilus_search_engine_zeitgeist_init (NautilusSearchEngineZeitgeist *engine)
{
        engine->details = g_new0 (NautilusSearchEngineZeitgeistDetails, 1);
}


NautilusSearchEngine *
nautilus_search_engine_zeitgeist_new (void)
{
        NautilusSearchEngineZeitgeist *engine;

        engine = g_object_new (NAUTILUS_TYPE_SEARCH_ENGINE_ZEITGEIST, NULL);
        
        engine->details->query_pending = FALSE;
        engine->details->cancellable = g_cancellable_new ();
        /*g_cancellable_connect (engine->details->cancellable,
                               G_CALLBACK (search_thread_done_idle),
                               engine,
                               NULL);*/

        return NAUTILUS_SEARCH_ENGINE (engine);
}
