#include "clutter-cover-flow-internal.h"


void
item_clear_behavior (CoverFlowItem *item, gpointer user_data)
{
    g_return_if_fail(item != NULL);

    if ( item->rotateBehaviour && 
         CLUTTER_IS_BEHAVIOUR(item->rotateBehaviour) && 
         clutter_behaviour_is_applied(item->rotateBehaviour, item->container) )
    {
        clutter_behaviour_remove(item->rotateBehaviour,item->container);
    }
}

void
item_free_visible(CoverFlowItem *item)
{
    /* Remove all item resources except the GFile, in case the
     * item is paged back in future */
    g_free(item->display_name);
    item->display_name = NULL;
    g_free(item->display_type);
    item->display_type = NULL;

    /* Remove and free any pending rotation behaviors */
    if (item->rotateBehaviour) {
        item_clear_behavior(item, NULL);
        item->rotateBehaviour = NULL;
    }

    if (item->container)
        clutter_actor_destroy(item->container);
    /* clutter_actor_destroy remove all childs if it s a container 
    ** etc so below is not really necessary */
#if 0
    if (item->container) {
        ClutterActor *parent;

        /* Remove the children actors first, they are unrefed automatically,
         * so should be collected */
        clutter_container_remove(
                                 CLUTTER_CONTAINER(item->container),
                                 item->texture,
                                 item->reflection,
                                 NULL);

        /* Remove ourselves from our parent */
        parent = clutter_actor_get_parent(item->container);
        clutter_container_remove_actor(
                                       CLUTTER_CONTAINER(parent),
                                       item->container);

        item->container = NULL;
    }
#endif
}

void 
item_free_invisible(CoverFlowItem *item)
{
    if(item!=NULL)
    {
        item_free_visible(item);
        g_object_unref(item->file);
        g_free(item);
    }
}

static void
get_item_count(ClutterCoverFlowPrivate *priv, NautilusFile *file)
{
    guint count;
    gboolean unreadable;

    //printf("mmm: %s\n", g_file_get_uri(file));
    if (nautilus_file_get_directory_item_count(nautilus_file_get_parent(file), &count, &unreadable))
        priv->m_item_count = count;
}

static NautilusFile
*get_nautilusfile_from_iter(ClutterCoverFlowPrivate *priv, GtkTreeIter *iter)
{
    NautilusFile *nfile;

    gtk_tree_model_get(GTK_TREE_MODEL(priv->model), iter, FM_LIST_MODEL_FILE_COLUMN, &nfile, -1 );
    g_return_val_if_fail (nfile, NULL);
    return (nfile);
}

CoverFlowItem *item_new(ClutterCoverFlowPrivate *priv, NautilusFile *nfile)
{
    if (nfile == NULL)
        return NULL;
    gpointer cb = NULL;
    CoverFlowItem *item;

    /* set a default info callback if one is not given */
    //if (NAUTILUS_IS_FILE(nfile))
    cb = g_object_get_qdata( G_OBJECT(nfile), priv->info_quark );
    if (cb == NULL)
        g_object_set_qdata( G_OBJECT(nfile), priv->info_quark, (gpointer)priv->get_info_callback );
    

    /* Create the new item */
    item = g_new0 (CoverFlowItem, 1);
    //item->get_info_callback = cb;
    item->file = nfile;
    item->get_info_callback = g_object_get_qdata( G_OBJECT(nfile), priv->info_quark );

    return (item);
}

static CoverFlowItem 
*item_list_grab_index(ClutterCoverFlowPrivate *priv, int idx)
{
    GtkTreePath *path;
    GtkTreeIter iter;

    if (idx >= 0)
    {
        path = gtk_tree_path_new_from_indices(idx, -1);
        if (gtk_tree_model_get_iter (priv->model, &iter, path))
        {
            gtk_tree_path_free (path);
            return (item_new(priv, get_nautilusfile_from_iter(priv, &iter)));
        }
        gtk_tree_path_free (path);
    }
    return NULL;
}

static CoverFlowItem
*item_icon_grab_index(ClutterCoverFlowPrivate *priv, int n)
{
    GList *p;
    char *uri;
    NautilusFile *file;

    p = g_list_nth(priv->icons, n);
    if (p != NULL)
    {
        uri = nautilus_icon_container_get_icon_uri(priv->container, p->data);
        if (uri)
        {
            file = nautilus_file_get_by_uri (uri);
            g_free(uri);
        }
        return (item_new(priv, file));
    }
    return NULL;
}

CoverFlowItem *item_grab_index(ClutterCoverFlowPrivate *priv, int idx)
{
    CoverFlowItem *item;
    
    if (priv->model)
        return (item_list_grab_index(priv, idx));
    else
        if (priv->container)
            return (item_icon_grab_index(priv, idx));
    return NULL;
}

static gboolean
foreach_func (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer user_data)
{
    CoverFlowItem *item;
    ClutterCoverFlowPrivate *priv;

    priv = user_data;
    //gtk_tree_model_get (model, iter, priv->file_column, &file, -1);
    gint *idxs = gtk_tree_path_get_indices(path);
    //printf("ADDED: %s %d\n", g_file_get_uri(item->file), idxs[0]);

    item = item_new(priv, get_nautilusfile_from_iter(priv, iter));
    if (item != NULL)
    {
        //printf("ADDED: %s %d dist:%d index:%d\n", nautilus_file_get_uri(item->file), idxs[0], idxs[0], idxs[0]+VISIBLE_ITEMS/2);
        priv->visible_items[idxs[0]+VISIBLE_ITEMS/2] = item;
        /* let appear all items in position 0 to get a nice first animation */
        view_add_item(priv, item, 0);
        animate_item_to_new_position(priv, item, idxs[0], MOVE_LEFT);
    }

    if (idxs[0] >= VISIBLE_ITEMS/2)
        return TRUE;
    return FALSE;
}

static void
fill_visible_items (GList *list, ClutterCoverFlowPrivate *priv)
{
    CoverFlowItem *item;
    GList *p;
    int index;
    char *uri;
    NautilusFile *file;

    for(p=list, index=0; p!=NULL && index <= VISIBLE_ITEMS/2; p=p->next, index++)
    {
        //printf("testal: %s\n", nautilus_icon_container_get_icon_uri(container, p->data));
        //printf("%s::%s %d\n", G_STRFUNC, nautilus_file_get_uri(list->data), index);	

      	uri = nautilus_icon_container_get_icon_uri(priv->container, p->data);
        //printf("%s::%s %d\n", G_STRFUNC, uri, index);

        file = nautilus_file_get_by_uri (uri);
        g_free (uri);

        item = item_new(priv, file);
        if (item != NULL)
        {
            //printf("ADDED: %s %d index:%d\n", nautilus_file_get_uri(item->file), index, index+VISIBLE_ITEMS/2);
            priv->visible_items[index+VISIBLE_ITEMS/2] = item;
            /* let appear all items in position 0 to get a nice first animation */
            view_add_item(priv, item, 0);
            animate_item_to_new_position(priv, item, index, MOVE_LEFT);
        }
    }
}

/* Return the indice of the tab where it match otherwise return -1 */
static int
is_incfitems(CoverFlowItem *item, CoverFlowItem **t)
{
    int i;
    for (i=0; i<VISIBLE_ITEMS; i++)
    {
        if (t[i])
        {
            //printf ("STRCMP[%d] %s %s\n", i, g_file_get_uri(priv->onstage_items[i]->file), g_file_get_uri(item->file));
            //if (!strcmp(nautilus_file_get_uri(t[i]->file), nautilus_file_get_uri(item->file)))
            if (t[i]->file == item->file)
                return i;
        }
    }
    return -1;
}

static void 
duplicate_visible_items(ClutterCoverFlowPrivate *priv)
{
    int i;

    for (i=0; i<VISIBLE_ITEMS; i++)
        priv->onstage_items[i] = priv->visible_items[i];
}

static void
remove_outof_range_actors(ClutterCoverFlowPrivate *priv)
{
    int i;

    for (i=0; i<VISIBLE_ITEMS; i++)
        if (priv->onstage_items[i])
        {
            if (is_incfitems(priv->onstage_items[i], priv->visible_items) == -1)
            {
                //printf ("delete: %s\n", nautilus_file_get_uri(priv->onstage_items[i]->file));
                item_free_invisible(priv->onstage_items[i]);
            }
        }
}
static void
get_selection_foreach_func (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data)
{
        //printf("%s -- path: %s\n", G_STRFUNC, gtk_tree_path_to_string(path));
}

void
items_update(ClutterCoverFlowPrivate *priv)
{
    GtkTreePath *path;
    CoverFlowItem *item;
    gint *idxs;
    int i, n;

    //g_message("%s", G_STRFUNC);

    GtkTreeSelection *selection = gtk_tree_view_get_selection(priv->tree);
    if (selection != NULL)
    {
        //printf ("got selection\n");
        //printf("nb rows selected %d\n", gtk_tree_selection_count_selected_rows(selection));
        if (!gtk_tree_selection_count_selected_rows(selection))
            priv->idx_visible_front = 0;
        else
        {
            GList *paths = gtk_tree_selection_get_selected_rows(selection, NULL);
            if (paths != NULL)
            {
                path = paths->data;
                idxs = gtk_tree_path_get_indices(path);
                //printf("finally got the path: %s\n", gtk_tree_path_to_string(path));
                //printf(":::: idx front %d selected %d\n", priv->idx_visible_front, idxs[0]);
                priv->idx_visible_front = idxs[0];
            }
        }
    }

    for (i=0; i<VISIBLE_ITEMS; i++)
    {
        item = item_grab_index(priv, priv->idx_visible_front-VISIBLE_ITEMS/2+i);
        if (item != NULL)
        {
            if ((n = is_incfitems(item, priv->onstage_items)) == -1)
            {
                //printf("ADDED: %s %d dist:%d index:%d\n", nautilus_file_get_uri(item->file), priv->idx_visible_front-VISIBLE_ITEMS/2+i, i-VISIBLE_ITEMS/2, i);
                view_add_item(priv, item, i-VISIBLE_ITEMS/2);
            }
            else
            {
                //printf("REUSE: %s %d dist:%d index:%d\n", nautilus_file_get_uri(item->file), priv->idx_visible_front-VISIBLE_ITEMS/2+i, i-VISIBLE_ITEMS/2, i);
                //item = priv->onstage_items[priv->idx_visible_front+VISIBLE_ITEMS/2];
                item = priv->onstage_items[n];
            }
            animate_item_to_new_position(priv, item, i-VISIBLE_ITEMS/2, MOVE_LEFT);
        }
        priv->visible_items[i] = item;
    }

    remove_outof_range_actors(priv);
    duplicate_visible_items(priv);
    view_restack(priv);
}

void
items_icon_update(ClutterCoverFlowPrivate *priv)
{
    GList *selection;

    g_return_if_fail (NAUTILUS_IS_ICON_CONTAINER (priv->container));
    //g_message("%s", G_STRFUNC);
	
    selection = nautilus_icon_container_get_selection (priv->container);
    NautilusIcon *icon;
    int index = 0;
    if (selection != NULL)
    {
       	icon = g_hash_table_lookup (priv->container->details->icon_set, selection->data);
	    if (icon) {
        	index = g_list_index (priv->icons, icon);    	
	    }
        //printf("::%s %d\n", G_STRFUNC, index);
        g_list_free (selection);
    }

    CoverFlowItem *item;
    int i, n;

    //printf(":::: idx front %d selected %d\n", priv->idx_visible_front, index);
    priv->idx_visible_front = index;
        

    for (i=0; i<VISIBLE_ITEMS; i++)
    {
        //item = item_icon_grab_index(priv, index-VISIBLE_ITEMS/2+i);
        item = item_grab_index(priv, index-VISIBLE_ITEMS/2+i);
        if (item != NULL)
        {
            if ((n = is_incfitems(item, priv->onstage_items)) == -1)
            {
                //printf("ADDED: %s %d dist:%d index:%d\n", nautilus_file_get_uri(item->file), index-VISIBLE_ITEMS/2+i, i-VISIBLE_ITEMS/2, i);
                view_add_item(priv, item, i-VISIBLE_ITEMS/2);
            }
            else
            {
                //printf("REUSE: %s %d dist:%d index:%d\n", nautilus_file_get_uri(item->file), index-VISIBLE_ITEMS/2+i, i-VISIBLE_ITEMS/2, i);
                //item = priv->onstage_items[index+VISIBLE_ITEMS/2];
                item = priv->onstage_items[n];
            }
            animate_item_to_new_position(priv, item, i-VISIBLE_ITEMS/2, MOVE_LEFT);
        }
        priv->visible_items[i] = item;
    }
  
    remove_outof_range_actors(priv);
    duplicate_visible_items(priv);
    view_restack(priv);
}

static void
model_do_insert(ClutterCoverFlowPrivate *priv, GtkTreeIter *iter, GtkTreePath *path, NautilusFile *file)
{
    /* get the number of files of current dir */
    get_item_count(priv, file);

    int nbs = model_get_length(priv);

    /* object appear inverted, not in good oder and gtk_tree_path_get_indices
    ** return incomprehensive results so we need to wait all the
    files before processing */
    //printf(":: nbs: %d m_item_count: %d\n", nbs, priv->m_item_count);
    //int titi = fm_list_model_get_length(priv->model);
    if (nbs == priv->m_item_count)
    {
        //printf("ini test nbs: %d m_item_count: %d\n", nbs, priv->m_item_count);
        if (!priv->visible_items)
        {
            priv->visible_items = g_new0 (CoverFlowItem*, VISIBLE_ITEMS);
            gtk_tree_model_foreach(priv->model, foreach_func, priv);
            duplicate_visible_items(priv);
            //printf("front fixed -- path: %s\n", gtk_tree_path_to_string(path));
        }
        else
        {
            items_update(priv);
        }
        update_item_text(priv, priv->visible_items[VISIBLE_ITEMS/2]);
    }
}

static NautilusFile *
model_get_file(ClutterCoverFlowPrivate *priv, GtkTreeIter *iter)
{
    NautilusFile *nfile;
    gtk_tree_model_get(GTK_TREE_MODEL(priv->model), iter, FM_LIST_MODEL_FILE_COLUMN, &nfile, -1 );
    return (nfile);
}

int
model_get_length(ClutterCoverFlowPrivate *priv)
{
    return gtk_tree_model_iter_n_children ( GTK_TREE_MODEL(priv->model), NULL );
}

int
model_get_position(ClutterCoverFlowPrivate *priv, GtkTreeIter *iter)
{
    GtkTreePath *path;
    gint *idx;

    path = gtk_tree_model_get_path( GTK_TREE_MODEL(priv->model), iter );
    idx = gtk_tree_path_get_indices(path);
    gtk_tree_path_free(path);

    return idx[0];    
}

gboolean
model_is_empty(ClutterCoverFlowPrivate *priv)
{
    if (priv->model)
        return model_get_length(priv) <= 0;
    else
        if (priv->container)
            return g_list_length (priv->container->details->icons) <= 0;
    return 1;
}

NautilusFile *
model_get_front_file(ClutterCoverFlowPrivate *priv)
{
    CoverFlowItem *item;
    
    g_critical("TODO: %s front %d", G_STRFUNC, priv->idx_visible_front);
    item = item_grab_index(priv, priv->idx_visible_front);
    if (item)
        return (item->file);
    return NULL;
}


void
model_row_loaded(GtkTreeModel *model, ClutterCoverFlowPrivate *priv)
{
    GtkTreePath *path;

    //printf("!!!%s\n", G_STRFUNC);
    if (!priv->visible_items)
    {
        priv->visible_items = g_new0 (CoverFlowItem*, VISIBLE_ITEMS);
        items_update(priv);

        //gtk_tree_model_foreach(priv->model, foreach_func, priv);
        //duplicate_visible_items(priv);
    }
}

void
model_row_inserted(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, ClutterCoverFlowPrivate *priv)
{
    NautilusFile *file;

    file = model_get_file(priv, iter);
    if (file) {
        model_do_insert(priv, iter, path, file);
        //g_message("GG: %s", nautilus_file_get_uri(file));
    }
}

void
model_row_changed(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, ClutterCoverFlowPrivate *priv)
{
    /* If the user adds a file to the model using the following idiom
       gtk_list_store_append (...);
       gtk_list_store_set (...);

       then we will get signals in the following order
       row-inserted (file = NULL)
       row-changed (file valid, not yet added)
       row-deleted (on old file)
       */

    //g_message("Changed: %s", G_STRFUNC);

#if 0
    NautilusFile *file;
    file = model_get_file(priv, iter);

    g_return_if_fail(file != NULL);

    if ( NULL == g_hash_table_lookup(priv->iter_added, iter) )
        model_do_insert(priv, iter, path, file);
    else
        g_critical("TODO: %s UPDATE: %s", G_STRFUNC, nautilus_file_get_uri(file));
#endif

}

void
model_row_reordered(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer arg3, ClutterCoverFlowPrivate *priv)
{
    /*printf("REORDERED path: %s\n", gtk_tree_path_to_string(path));
    if (path == NULL)
        printf("FAILED\n");*/
    items_update(priv);
}

void
model_row_deleted(GtkTreeModel *model, GtkTreePath *path, ClutterCoverFlowPrivate *priv)
{
    g_critical("TODO: %s", G_STRFUNC);
}

/* Icon View Events */
static void
testal_print(NautilusIconContainer *container)
{
    GList *p;

    for(p=container->details->icons; p!=NULL; p=p->next)
    {
        printf("testal: %s\n", nautilus_icon_container_get_icon_uri(container, p->data));
    }
}

void
model_icon_added(NautilusIconContainer *container, NautilusIconData *icon_data, ClutterCoverFlowPrivate *priv)
{
    //g_message("%s", G_STRFUNC);
    //printf("::%s\n", nautilus_icon_container_get_icon_uri(container, icon_data));
    //testal_print(container);
      
    if (container->details->icons)
    {
        priv->icons = container->details->icons;
        if (!priv->visible_items)
        {
        /*NautilusDirectory *ndir = nautilus_directory_get_for_file((NautilusFile *)icon_data);
        printf("ndir: %s\n", nautilus_directory_get_uri(ndir));
        priv->icons = nautilus_directory_get_file_list(ndir);*/

            priv->visible_items = g_new0 (CoverFlowItem*, VISIBLE_ITEMS);
            fill_visible_items(priv->icons, priv);
            duplicate_visible_items(priv);
        }
        else
        {
            items_icon_update(priv);
        }
        update_item_text(priv, priv->visible_items[VISIBLE_ITEMS/2]);
    }
}

void
model_icon_changed(NautilusIconContainer *container, NautilusIconData *icon_data, gpointer data)
{
    g_message("%s", G_STRFUNC);
}

void
model_icon_text_changed(NautilusIconContainer *container, NautilusIconData *icon_data, gpointer data)
{
    g_message("%s", G_STRFUNC);
}

void
model_icon_reordered(NautilusIconContainer *container, gpointer data)
{
    //amtest no in the good function
    /* 
    NautilusFile *file = fm_directory_view_get_directory_as_file (view);
    char *sort_name = fm_icon_view_get_directory_sort_by (icon_view, file);
    get_sort_criterion_by_metadata_text (sort_name);
    fm_icon_view_get_directory_sort_reversed (icon_view, file);
    */
    g_message("%s", G_STRFUNC);
}

void
model_icon_removed(NautilusIconContainer *container, NautilusIconData *icon_data, gpointer data)
{
    g_message("%s", G_STRFUNC);
}

void
model_icon_selection_changed(NautilusIconContainer *container, gpointer data)
{
    GList *selection, *p;
    NautilusIcon *icon;
    ClutterCoverFlow *cf;
    ClutterCoverFlowPrivate *priv;

    g_return_if_fail (NAUTILUS_IS_ICON_CONTAINER (container));
    //g_message("%s", G_STRFUNC);
	
    cf = data;
    priv = cf->priv;
    selection = nautilus_icon_container_get_selection (container);
   
    if (container->details->icons)
        priv->icons = container->details->icons;

    int index = 0;
    if (selection != NULL)
    {
       	icon = g_hash_table_lookup (priv->container->details->icon_set, selection->data);
	    if (icon) {
        	index = g_list_index (priv->icons, icon);    	
        	//index = g_list_index (container->details->icons, icon);    	
	    }
        //printf("::index first selected item %d\n", index);
    }
    clutter_cover_flow_scroll_to_position(cf, index);

    /*for (p = selection; p != NULL; p = p->next)
    {
        printf("selection : %s\n", nautilus_file_get_uri(p->data));
    }*/
   	//nautilus_file_list_free (selection);
    g_list_free (selection);
}

void
view_restack(ClutterCoverFlowPrivate *priv)
{
    int i;
    CoverFlowItem *j, *k;// *l;
    
    //g_message("Restack");
    g_return_if_fail(priv->visible_items);
    /*Restack items in the left*/
    for(i = 0; i< VISIBLE_ITEMS/2; i++)
    {
        j = priv->visible_items[i];
        k = priv->visible_items[i+1];
        if(k != NULL && j != NULL)
            clutter_actor_raise(k->container, j->container);
    }   
    /*Restack items in the Right*/
    for(i = VISIBLE_ITEMS-1; i> VISIBLE_ITEMS/2+1; i--)
    {
        j = priv->visible_items[i];
        k = priv->visible_items[i-1];
        if(k != NULL && j != NULL)
            clutter_actor_raise(k->container, j->container);
    }   

    /*Raise cover*/
    if (k = priv->visible_items[VISIBLE_ITEMS/2])
    {
        clutter_actor_raise_top(k->container);
        clutter_actor_raise_top(priv->item_name);
        clutter_actor_raise_top(priv->item_type);
    }
}


void
view_add_item(ClutterCoverFlowPrivate *priv, CoverFlowItem *item, int pos)
{
    int bps;
    float scale, dist;
    GdkPixbuf *pb;

    g_return_if_fail(item != NULL);
    g_return_if_fail(item->file != NULL);
    g_return_if_fail(item->get_info_callback != NULL);

    //g_return_if_fail(n_visible_items < VISIBLE_ITEMS);

    //g_message("ADD AT POS: %d FRONT %d", pos, priv->idx_visible_front);

    /* Get the information about the item */
    item->get_info_callback(
                            item->file,
                            &(item->display_name),
                            &(item->display_type),
                            &pb,
                            DEFAULT_ICON_SIZE);

    item->texture = black_texture_new();
    if( gdk_pixbuf_get_has_alpha(pb) )
        bps = 4;
    else
        bps = 3;

    clutter_texture_set_from_rgb_data(
                                      CLUTTER_TEXTURE(item->texture), 
                                      gdk_pixbuf_get_pixels       (pb),
                                      gdk_pixbuf_get_has_alpha    (pb),
                                      gdk_pixbuf_get_width        (pb),
                                      gdk_pixbuf_get_height       (pb),
                                      gdk_pixbuf_get_rowstride    (pb),
                                      bps,
                                      (ClutterTextureFlags)0,
                                      NULL);

    scale_to_fit (item->texture);

    /* Reflection */
    //item->reflection = clutter_clone_new ( item->texture );
    item->reflection = black_texture_new();

    clutter_texture_set_from_rgb_data(
                                      CLUTTER_TEXTURE(item->reflection), 
                                      gdk_pixbuf_get_pixels       (pb),
                                      gdk_pixbuf_get_has_alpha    (pb),
                                      gdk_pixbuf_get_width        (pb),
                                      gdk_pixbuf_get_height       (pb),
                                      gdk_pixbuf_get_rowstride    (pb),
                                      bps,
                                      (ClutterTextureFlags)0,
                                      NULL);

    g_object_unref(pb);

    scale_to_fit ( item->reflection );
    clutter_actor_set_position ( item->reflection, 0, clutter_actor_get_height(item->reflection) );
    clutter_actor_set_rotation (
                                item->reflection,
                                CLUTTER_Z_AXIS,180,
                                clutter_actor_get_width(item->reflection)/2,
                                clutter_actor_get_height(item->reflection)/2,
                                0);
    clutter_actor_set_rotation (
                                item->reflection,
                                CLUTTER_Y_AXIS,180,
                                clutter_actor_get_width(item->reflection)/2,
                                clutter_actor_get_height(item->reflection)/2,
                                0);

    /* Container */
    item->container = clutter_group_new();
    clutter_container_add (
                           CLUTTER_CONTAINER(item->container),
                           item->texture,
                           item->reflection,
                           NULL);
    clutter_container_add_actor (
                                 CLUTTER_CONTAINER(priv->m_container),
                                 item->container);

    scale = get_item_scale(item, pos);
    dist = get_item_distance(item, pos);

    clutter_actor_set_position ( 
                                item->container, 
                                dist -clutter_actor_get_width(item->texture)/2, 
                                -clutter_actor_get_height(item->texture)/2);

    clutter_actor_set_scale_full (
                                  item->container, scale, scale,
                                  clutter_actor_get_width(item->texture)/2,
                                  clutter_actor_get_height(item->texture)/2);
    /* Animate the fade in of the new item */
    fade_in (priv, item, pos);
}

void
view_move(ClutterCoverFlowPrivate *priv, move_t dir, gboolean move_ends)
{

    int i, dist;   
    CoverFlowItem *item;

    if (!priv->visible_items)
        return;
    if(dir == MOVE_LEFT /*&&  priv->idx_visible_front NOT AT THE END*/) /*not at ***O position*/
    {
        item_free_invisible(priv->visible_items[0]);    /*Free first item*/
        priv->idx_visible_front++;                      /*Update front cover idx*/

        /*     xxOxx ===> xxxOx       */
        for(i=0; i<VISIBLE_ITEMS-1; i++)
        {
            priv->visible_items[i] = priv->visible_items[i+1]; 
            //dist = view_calc_dist_from_front(priv,i) - 3;
            dist = i - VISIBLE_ITEMS/2;

            /*If there is a cover*/        
            if(priv->visible_items[i] != NULL)
            {
                //g_message("Moving left %s %i front cover is %i so dist is %i", priv->visible_items[i]->display_name, i , VISIBLE_ITEMS/2, dist);
                animate_item_to_new_position(priv, priv->visible_items[i], dist , dir);
            }
        }
        /* Add the new item*/
        //priv->visible_items[VISIBLE_ITEMS-1] = NULL;
        item = item_grab_index(priv, priv->idx_visible_front + VISIBLE_ITEMS/2);
        priv->visible_items[VISIBLE_ITEMS-1] = item;
        if (item != NULL)
        {
            view_add_item(priv, item, VISIBLE_ITEMS/2);
            animate_item_to_new_position(priv, item, VISIBLE_ITEMS/2, MOVE_LEFT);
        }
        duplicate_visible_items(priv);
    }
    /*TODO: ADD new item and animate it*/
    //if(dir == MOVE_RIGHT &&  priv->idx_visible_front > 0)   /*not at O*** position*/
    if(dir == MOVE_RIGHT)   /*not at O*** position*/
    {
        /*Free last item*/
        item_free_invisible(priv->visible_items[VISIBLE_ITEMS-1]);
        priv->idx_visible_front--;


        for(i=VISIBLE_ITEMS-1; i>0; i--)
        {
            priv->visible_items[i] = priv->visible_items[i-1];
            dist = i - VISIBLE_ITEMS/2;

            /*If there is a cover*/        
            if(priv->visible_items[i] != NULL)
            {
                //g_message("Moving left %s %i front cover is %i so dist is %i", priv->visible_items[i]->display_name, i , VISIBLE_ITEMS/2, dist);
                animate_item_to_new_position(priv, priv->visible_items[i], dist, dir);
            }
        } 
        /* Add the new item*/
        //priv->visible_items[0] = NULL;      
        item = item_grab_index(priv, priv->idx_visible_front - VISIBLE_ITEMS/2);
        priv->visible_items[0] = item;
        if (item != NULL)
        {
            view_add_item(priv, item, -VISIBLE_ITEMS/2);
            animate_item_to_new_position(priv, item, -VISIBLE_ITEMS/2, MOVE_LEFT);
        }
        duplicate_visible_items(priv);
    }

    view_restack(priv);
    
    /*Update text*/
    update_item_text(priv, priv->visible_items[VISIBLE_ITEMS/2]);
}


/* amtest: usefull? */
void
reset(ClutterCoverFlowPrivate *priv)
{
    ClutterActor *stage;
    guint h,w;
    float relation;

    stage = clutter_actor_get_stage(priv->m_container);

    h = clutter_actor_get_height(stage);
    w = clutter_actor_get_width(stage);
    relation = (float)500/(float)h;

    clutter_actor_set_position(
                               priv->m_container,
                               w/2,
                               h/2);

    clutter_actor_set_position (
                                priv->item_name,
                                0 - clutter_actor_get_width(priv->item_name)/2,
                                (MAX_ITEM_HEIGHT/2) + TEXT_PAD_BELOW_ITEM);
    clutter_actor_set_position (
                                priv->item_type,
                                0 - clutter_actor_get_width(priv->item_type)/2,
                                (MAX_ITEM_HEIGHT/2) + TEXT_PAD_BELOW_ITEM + clutter_actor_get_height(priv->item_name) + 5);

    clutter_actor_set_scale(
                            priv->m_container,
                            1/relation,
                            1/relation);

    /*clutter_text_set_text(
                          CLUTTER_TEXT(priv->item_name),
                          "");*/
    clutter_actor_set_opacity(
                              priv->item_name,
                              255);

    /*clutter_text_set_text(
                          CLUTTER_TEXT(priv->item_type),
                          "");*/
    clutter_actor_set_opacity(
                              priv->item_type,
                              255);

}

void
zoom_items(ClutterCoverFlowPrivate *priv, float zoom_value, gboolean clear_when_complete)
{
#if 0
    GSequenceIter *iter;
    ClutterAnimation *anim;
    ClutterAlpha *alpha;
    ClutterTimeline *timeline;

    timeline = clutter_timeline_new(500);
    alpha = clutter_alpha_new_full(timeline, CLUTTER_EASE_OUT_EXPO);

    /* scale the container */
    anim = clutter_actor_animate_with_alpha (
                                             priv->m_container,
                                             alpha,
                                             "scale-x", zoom_value,
                                             "scale-y", zoom_value,
                                             NULL);

    /* shade out all the items */
    for (iter = priv->iter_visible_start;
         TRUE;
         iter = g_sequence_iter_next(iter))
    {
        CoverFlowItem *item;
        item = g_sequence_get(iter);
        clutter_actor_animate_with_alpha (
                                          item->texture,
                                          alpha,
                                          "shade", 0,
                                          NULL);
        clutter_actor_animate_with_alpha (
                                          item->reflection,
                                          alpha,
                                          "shade", 0,
                                          NULL);
        clutter_timeline_stop(timeline);
        if (iter == priv->iter_visible_end)
            break;
    }

    if (clear_when_complete) {
        g_signal_connect (
                          timeline, "completed",
                          G_CALLBACK (items_free_all), priv);
    } else {
        g_signal_connect (
                          timeline, "completed",
                          G_CALLBACK (items_show_all), priv);
    }

    clutter_timeline_start(timeline);
#endif
}

void
knock_down_items(ClutterCoverFlowPrivate *priv, gboolean clear_when_complete)
{
#if 0
    GSequenceIter *iter;
    ClutterTimeline *timeline;
    ClutterAlpha *alpha;

    timeline = clutter_timeline_new(500);
    alpha = clutter_alpha_new_full (timeline,CLUTTER_EASE_OUT_EXPO);

    for (iter = priv->iter_visible_start;
         TRUE;
         iter = g_sequence_iter_next(iter))
    {
        ClutterBehaviour *up, *down;
        CoverFlowItem *item;

        item = g_sequence_get(iter);

        down = clutter_behaviour_rotate_new (
                                             alpha,
                                             CLUTTER_X_AXIS,
                                             CLUTTER_ROTATE_CCW,
                                             clutter_actor_get_rotation(item->texture,CLUTTER_Y_AXIS,0,0,0),
                                             270);
        clutter_behaviour_rotate_set_center ( 
                                             CLUTTER_BEHAVIOUR_ROTATE(down),
                                             0,
                                             clutter_actor_get_height(item->texture),
                                             0);
        clutter_behaviour_apply (down, item->texture);

        up = clutter_behaviour_rotate_new (
                                           alpha,
                                           CLUTTER_X_AXIS,
                                           CLUTTER_ROTATE_CW,
                                           clutter_actor_get_rotation(item->texture,CLUTTER_Y_AXIS,0,0,0),
                                           360-270);  
        clutter_behaviour_rotate_set_center ( 
                                             CLUTTER_BEHAVIOUR_ROTATE(up),
                                             0,
                                             clutter_actor_get_height(item->reflection),
                                             0);
        clutter_behaviour_apply (up, item->reflection);

        /* Animate with alpha starts the timeline */
        clutter_actor_animate_with_alpha (
                                          item->texture,
                                          alpha,
                                          "shade", 0,
                                          NULL);
        clutter_actor_animate_with_alpha (
                                          item->reflection,
                                          alpha,
                                          "shade", 0,
                                          NULL);
        clutter_timeline_stop(timeline);
        if (iter == priv->iter_visible_end)
            break;
    }

    /* Also fade out the text */
    clutter_actor_animate_with_alpha (
                                      priv->item_name,
                                      alpha,
                                      "opacity", 0,
                                      NULL);
    clutter_actor_animate_with_alpha (
                                      priv->item_type,
                                      alpha,
                                      "opacity", 0,
                                      NULL);

    if (clear_when_complete) {
        g_signal_connect (
                          timeline, "completed",
                          G_CALLBACK (items_free_all), priv);
    } else {
        g_signal_connect (
                          timeline, "completed",
                          G_CALLBACK (items_show_all), priv);
    }

    clutter_timeline_start(timeline);
#endif
}

int
get_actor_pos(ClutterCoverFlowPrivate *priv, ClutterActor * actor)
{
    int i;

    for (i=0; i<VISIBLE_ITEMS; i++)
    {
        if (priv->visible_items[i])
            if (priv->visible_items[i]->texture == actor)
                return i-VISIBLE_ITEMS/2;
    }
    return 0;
}

/*
 * This functions adds a rotation behaviour from the current angle to the final angle 
 * rotating with the direction <direction> 
 */
void
set_rotation_behaviour (ClutterCoverFlowPrivate *priv, CoverFlowItem *item, int final_angle, ClutterRotateDirection direction)
{
    double current;

    current = clutter_actor_get_rotation(item->container,CLUTTER_Y_AXIS,0,0,0);
    if(current<0) current+=360;
    if(current>360) current-=360;

    if(current != final_angle) {
        item->rotateBehaviour = clutter_behaviour_rotate_new (
                                                              priv->m_alpha,
                                                              CLUTTER_Y_AXIS,
                                                              direction,
                                                              current,
                                                              final_angle);
        clutter_behaviour_rotate_set_center ( 
                                             CLUTTER_BEHAVIOUR_ROTATE(item->rotateBehaviour),
                                             clutter_actor_get_width(item->container)/2,
                                             0,0);
        clutter_behaviour_apply (item->rotateBehaviour, item->container);
    }
}

float
get_item_scale(CoverFlowItem *item, int dist_from_front)
{
    if(dist_from_front == 0 )
        return MAX_SCALE;
    else
        return 1.0;
}

void
get_item_angle_and_dir(CoverFlowItem *item, int dist_from_front, move_t dir, int *angle, ClutterRotateDirection *rotation_dir)
{
    /* The front item direction depends on the direction we came from */
    if (dist_from_front == 0) {
        *rotation_dir =  (dir == MOVE_RIGHT ? CLUTTER_ROTATE_CCW : CLUTTER_ROTATE_CW);
        *angle = 0;
    }

    /* Item on the right */
    if (dist_from_front > 0) {
        *rotation_dir = CLUTTER_ROTATE_CCW;
        *angle = 360 - MAX_ANGLE;
    }

    /* Item on the left */
    if (dist_from_front < 0) {
        *rotation_dir = CLUTTER_ROTATE_CW;
        *angle = MAX_ANGLE;
    }
}

gfloat
get_item_distance(CoverFlowItem *item, int dist_from_front)
{
    gfloat dist = (ABS(dist_from_front) * COVER_SPACE) + FRONT_COVER_SPACE;

    if (dist_from_front == 0)
        return 0.0;

    if (dist_from_front < 0)
        dist *= -1.0;

    return dist;
}

int
get_item_opacity(CoverFlowItem *item, int dist_from_front)
{
    //return CLAMP(255*(VISIBLE_ITEMS - ABS(dist_from_front))/VISIBLE_ITEMS, 0, 255);
    //return CLAMP(255*((VISIBLE_ITEMS/2)+1 - ABS(dist_from_front))/(VISIBLE_ITEMS/2), 0, 255);
    return CLAMP(255*((V_ITEMS/2)+1 - ABS(dist_from_front))/(V_ITEMS/2), 0, 255);
}

int
get_item_reflection_opacity(CoverFlowItem *item, int dist_from_front)
{
    //return CLAMP((REFLECTION_ALPHA*(VISIBLE_ITEMS - ABS(dist_from_front))/VISIBLE_ITEMS), 0, REFLECTION_ALPHA);
    //return CLAMP((REFLECTION_ALPHA*((VISIBLE_ITEMS/2)+1 - ABS(dist_from_front))/(VISIBLE_ITEMS/2)), 0, REFLECTION_ALPHA);
    return CLAMP((REFLECTION_ALPHA*((V_ITEMS/2)+1 - ABS(dist_from_front))/(V_ITEMS/2)), 0, REFLECTION_ALPHA);
}

void
animate_item_to_new_position(ClutterCoverFlowPrivate *priv, CoverFlowItem *item, int dist_from_front, move_t dir)
{
    g_return_if_fail (priv->m_alpha);
    g_return_if_fail (CLUTTER_IS_ALPHA (priv->m_alpha));
    float scale, dist;
    int angle, opacity, reflection_opacity;
    ClutterRotateDirection rotation_dir = 0;

    scale = get_item_scale(item, dist_from_front);
    dist = get_item_distance(item, dist_from_front);
    opacity = get_item_opacity(item, dist_from_front);
    reflection_opacity = get_item_reflection_opacity(item, dist_from_front);
    get_item_angle_and_dir(item, dist_from_front, dir, &angle, &rotation_dir);

    set_rotation_behaviour(priv, item, angle, rotation_dir);

    clutter_actor_animate_with_alpha (item->texture, priv->m_alpha,
                                      "shade", opacity,
                                      NULL);
    clutter_actor_animate_with_alpha (item->reflection, priv->m_alpha,
                                      "shade", reflection_opacity,
                                      NULL);

    clutter_actor_animate_with_alpha (
                                      item->container,
                                      priv->m_alpha,
                                      "scale-x", scale,
                                      "scale-y", scale,
                                      "scale-center-x" , clutter_actor_get_width  (item->texture)/2,
                                      "scale-center-y" , clutter_actor_get_height (item->texture)/2,
                                      "x", dist - clutter_actor_get_width(item->container)/2,
                                      NULL);
    
    clutter_actor_lower_bottom(item->container);
}

void
update_item_text(ClutterCoverFlowPrivate *priv, CoverFlowItem *item)
{
    g_return_if_fail(item);
    /* Set text in the centre to the name of the file */
    clutter_text_set_text(
                          CLUTTER_TEXT(priv->item_name),
                          item->display_name);
    clutter_actor_set_x(
                        priv->item_name, 
                        0 - clutter_actor_get_width(priv->item_name)/2);
    clutter_text_set_text(
                          CLUTTER_TEXT(priv->item_type),
                          item->display_type);
    clutter_actor_set_x(
                        priv->item_type, 
                        0 - clutter_actor_get_width(priv->item_type)/2);
}

void
start(ClutterCoverFlow *self)
{
    // useless with clutter_actor_animate
    //clutter_timeline_start(self->priv->m_timeline);
}

void
stop(ClutterCoverFlow *self)
{
    clutter_timeline_stop(self->priv->m_timeline);
}

void
clear_behaviours (ClutterCoverFlow *self)
{
    int i;
    ClutterCoverFlowPrivate *priv;

    priv = self->priv;
    g_return_if_fail(priv->visible_items);
    if ( ! model_is_empty(priv) ) 
        for (i=0; i<VISIBLE_ITEMS; i++)
            if (priv->visible_items[i])
                if (CLUTTER_IS_BEHAVIOUR(priv->visible_items[i]->rotateBehaviour) && 
                    clutter_behaviour_is_applied(priv->visible_items[i]->rotateBehaviour, priv->visible_items[i]->container) )
                {
                    clutter_behaviour_remove(CLUTTER_BEHAVIOUR(priv->visible_items[i]->rotateBehaviour), priv->visible_items[i]->container);
                }
}

void
fade_in(ClutterCoverFlowPrivate *priv, CoverFlowItem *item, guint distance_from_centre)
{
    int opacity = get_item_opacity(item, distance_from_centre);
    int reflection_opacity 	= get_item_reflection_opacity(item, distance_from_centre);

    clutter_actor_animate_with_alpha (item->texture, priv->m_alpha, "shade", opacity, NULL);
    clutter_actor_animate_with_alpha (item->reflection, priv->m_alpha , "shade", reflection_opacity, NULL);
}

void
scale_to_fit(ClutterActor *actor)
{
    int w = clutter_actor_get_width(actor);
    int h = clutter_actor_get_height(actor);

    if( h > MAX_ITEM_HEIGHT) {
        int temp = w*MAX_ITEM_HEIGHT/h;
        clutter_actor_set_size(actor, temp, MAX_ITEM_HEIGHT);
    }
}

void
get_info(GFile *file, char **name, char **description, GdkPixbuf **pb, guint pbsize)
{
    GIcon *icon;
    GtkIconInfo *icon_info;
    GFileInfo *file_info;
    GtkIconTheme *icon_theme;
    char *uri;

    uri = g_file_get_uri(file);
    g_debug("Default get info: %s", uri);
    g_free(uri);

    icon_theme = gtk_icon_theme_get_default();
    file_info = g_file_query_info(
                                  file,
                                  G_FILE_ATTRIBUTE_STANDARD_ICON "," G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME "," G_FILE_ATTRIBUTE_STANDARD_FAST_CONTENT_TYPE,
                                  G_FILE_QUERY_INFO_NONE, NULL , NULL);
    icon = g_file_info_get_icon(file_info);
    icon_info = gtk_icon_theme_lookup_by_gicon(
                                               icon_theme,
                                               icon,
                                               pbsize,
                                               GTK_ICON_LOOKUP_USE_BUILTIN | GTK_ICON_LOOKUP_GENERIC_FALLBACK | GTK_ICON_LOOKUP_FORCE_SIZE);

    *pb = gtk_icon_info_load_icon(icon_info, NULL);
    *name = g_strdup(
                     g_file_info_get_display_name(file_info));
    *description = g_content_type_get_description(
                                                  g_file_info_get_attribute_string(
                                                                                   file_info,
                                                                                   G_FILE_ATTRIBUTE_STANDARD_FAST_CONTENT_TYPE));

    g_object_unref(file_info);
    gtk_icon_info_free(icon_info);
}

