/* vim: set et ts=4 sw=4: */

#include <gtk/gtk.h>
#include <glib.h>
#include "clutter-cover-flow.h"
#include "clutter-cover-flow-internal.h"
#include "clutter-black-texture.h"

G_DEFINE_TYPE (ClutterCoverFlow, clutter_cover_flow, CLUTTER_TYPE_GROUP)

#define GET_PRIVATE(o) \
        (G_TYPE_INSTANCE_GET_PRIVATE ((o), CLUTTER_TYPE_COVER_FLOW, ClutterCoverFlowPrivate))

#define USE_LIBGNOME_FOR_THUMB	1

static void
clutter_cover_flow_dispose (GObject *object)
{
    ClutterCoverFlow *self = CLUTTER_COVER_FLOW(object); 

    if (self->priv)
    {
        //if (self->priv->trans != NULL)
        //g_object_unref(self->priv->trans);
        //self->priv->trans = NULL;
    }

    G_OBJECT_CLASS (clutter_cover_flow_parent_class)->dispose (object);
}

static void
clutter_cover_flow_finalize (GObject *object)
{
    ClutterCoverFlow *self = CLUTTER_COVER_FLOW(object); 

    if (self->priv)
    {
        g_free(self->priv);
        self->priv = NULL;
    }

    G_OBJECT_CLASS (clutter_cover_flow_parent_class)->finalize (object);
}

static void
clutter_cover_flow_class_init (ClutterCoverFlowClass *klass)
{
    GObjectClass *object_class;
    //ClutterActorClass *actor_class;

    object_class = G_OBJECT_CLASS (klass);

    object_class->dispose = clutter_cover_flow_dispose;
    object_class->finalize = clutter_cover_flow_finalize;
}

static void
clutter_cover_flow_init (ClutterCoverFlow *self)
{
    self->priv  = g_new0 (ClutterCoverFlowPrivate, 1);
    //self->priv->visible_items = g_new0 (CoverFlowItem*, VISIBLE_ITEMS);
    self->priv->onstage_items = g_new0 (CoverFlowItem*, VISIBLE_ITEMS);

    self->priv->m_timeline = clutter_timeline_new(FRAMES * FPS);
    self->priv->m_alpha = clutter_alpha_new_full(self->priv->m_timeline,CLUTTER_EASE_OUT_EXPO);

    self->priv->info_quark = g_quark_from_string("INFO_CALLBACK");
    self->priv->item_quark = g_quark_from_string("ITEM");

    self->priv->view_mode = COVERFLOW_MODE;
    self->priv->visible_items = 0;
}

static gboolean
on_stage_resized(ClutterStage *stage, ClutterButtonEvent *event, gpointer user_data)
{
    ClutterCoverFlow *self = CLUTTER_COVER_FLOW(user_data);
    reset(self->priv);
    return TRUE;
}

void 
clutter_cover_flow_set_model(ClutterCoverFlow *self, GtkTreeView *tree, GtkTreeModel *store)
{
    g_return_if_fail( CLUTTER_IS_COVER_FLOW(self) );

    g_return_if_fail( GTK_IS_TREE_MODEL(store) );
    self->priv->model = store;
    self->priv->container = NULL;
    self->priv->tree = tree;

    /*g_signal_connect (store, "row-loaded",
                      G_CALLBACK (model_row_loaded), (gpointer)(self->priv));*/
    g_signal_connect (store, "row-inserted",
                      G_CALLBACK (model_row_inserted), (gpointer)(self->priv));
    g_signal_connect (store, "row-changed",
                      G_CALLBACK (model_row_changed), (gpointer)(self->priv));
    g_signal_connect (store, "rows-reordered",
                      G_CALLBACK (model_row_reordered), (gpointer)(self->priv));
    g_signal_connect (store, "row-deleted",
                      G_CALLBACK (model_row_deleted), (gpointer)(self->priv));
}

void 
clutter_cover_flow_unset_model(ClutterCoverFlow *self, GtkTreeModel *store)
{
    g_return_if_fail( CLUTTER_IS_COVER_FLOW(self) );

    g_return_if_fail( GTK_IS_TREE_MODEL(store) );

    g_signal_handlers_disconnect_by_func (store,
                      G_CALLBACK (model_row_inserted), (gpointer)(self->priv));
    g_signal_handlers_disconnect_by_func (store,
                      G_CALLBACK (model_row_changed), (gpointer)(self->priv));
    g_signal_handlers_disconnect_by_func (store,
                      G_CALLBACK (model_row_reordered), (gpointer)(self->priv));
    g_signal_handlers_disconnect_by_func (store,
                      G_CALLBACK (model_row_deleted), (gpointer)(self->priv));
}

void 
clutter_cover_flow_set_model_icon(ClutterCoverFlow *self, NautilusIconContainer *container)
{
    g_return_if_fail( CLUTTER_IS_COVER_FLOW(self) );

    g_return_if_fail( NAUTILUS_IS_ICON_CONTAINER(container) );
    self->priv->container = container;
    self->priv->model = NULL;

    //g_signal_connect (container, "icon_added",
    g_signal_connect (container, "new_icons_added",
                      G_CALLBACK (model_icon_added), (gpointer)(self->priv));
    g_signal_connect (container, "icon_position_changed",
                      G_CALLBACK (model_icon_changed), (gpointer)(self->priv));
    g_signal_connect (container, "icon_text_changed",
                      G_CALLBACK (model_icon_text_changed), (gpointer)(self->priv));
    g_signal_connect (container, "layout_changed",
                      G_CALLBACK (model_icon_reordered), (gpointer)(self->priv));
    g_signal_connect (container, "icon_removed",
                      G_CALLBACK (model_icon_removed), (gpointer)(self->priv));
    g_signal_connect (container, "selection_changed",
                      G_CALLBACK (model_icon_selection_changed), (gpointer)(self));

}

void 
clutter_cover_flow_unset_model_icon(ClutterCoverFlow *self, NautilusIconContainer *container)
{
    g_return_if_fail( CLUTTER_IS_COVER_FLOW(self) );

    g_return_if_fail( NAUTILUS_IS_ICON_CONTAINER(container) );

    //g_signal_connect (container, "icon_added",
    g_signal_handlers_disconnect_by_func (container,
                      G_CALLBACK (model_icon_added), (gpointer)(self->priv));
    g_signal_handlers_disconnect_by_func (container,
                      G_CALLBACK (model_icon_changed), (gpointer)(self->priv));
    g_signal_handlers_disconnect_by_func (container,
                      G_CALLBACK (model_icon_text_changed), (gpointer)(self->priv));
    g_signal_handlers_disconnect_by_func (container,
                      G_CALLBACK (model_icon_reordered), (gpointer)(self->priv));
    g_signal_handlers_disconnect_by_func (container,
                      G_CALLBACK (model_icon_removed), (gpointer)(self->priv));
    g_signal_handlers_disconnect_by_func (container,
                      G_CALLBACK (model_icon_selection_changed), (gpointer)(self));

}

static void 
get_info_nautilus_thumb(NautilusFile *nfile, char **name, char **description, GdkPixbuf **pb, guint pbsize)
{
	char *uri;
	int thumb_flags = NAUTILUS_FILE_ICON_FLAGS_USE_THUMBNAILS | NAUTILUS_FILE_ICON_FLAGS_FORCE_THUMBNAIL_SIZE;

	uri = nautilus_file_get_uri(nfile);
	g_debug("Get info: %s", uri);
	g_free(uri);

	*pb = nautilus_file_get_icon_pixbuf (
		nfile, 
		pbsize, 
		TRUE, 
		thumb_flags);

	*name = nautilus_file_get_display_name(nfile);

	/* TODO: Perhaps I should call nautils_file_get_description, but
         * it gives up if the standard::description key has not been loaded */
	*description = nautilus_file_get_string_attribute(nfile, "type");

	//nautilus_file_unref(nfile);
}

/* 
** nautilus_file_get_icon_pixbuf is for icon where the largest value is 
** nautilus-icon-info.h:#define NAUTILUS_ICON_SIZE_LARGEST     192
** this results having images scaled from small 192 sized thumb
** real thumbnail that nautilus use, use fcts defined in
** ../../libnautilus-private/nautilus-thumbnails.c
**
** so let's use gnome_desktop_thumbnail's new experimental API
*/

static void 
get_info_libgnome_thumb(NautilusFile *nfile, char **name, char **description, GdkPixbuf **pb, guint pbsize)
{
    GnomeDesktopThumbnailFactory *fac;
    char *uri;
    time_t mtime;
    char *thumb_loc;

    fac = gnome_desktop_thumbnail_factory_new(GNOME_DESKTOP_THUMBNAIL_SIZE_LARGE);
    uri = nautilus_file_get_uri (nfile);
    mtime = nautilus_file_get_mtime (nfile);

    *name = nautilus_file_get_display_name(nfile);
    *description = nautilus_file_get_string_attribute(nfile, "type");

    thumb_loc = gnome_desktop_thumbnail_factory_lookup (fac, uri, mtime);
    if (!thumb_loc) 
    {
        if ((*pb = gnome_desktop_thumbnail_factory_generate_thumbnail(fac, uri, *description)))
            gnome_desktop_thumbnail_factory_save_thumbnail(fac, *pb, uri, mtime);
        else
            *pb =  nautilus_file_get_icon_pixbuf (nfile, pbsize, TRUE, 0);      
            //*pb =  nautilus_file_get_icon_pixbuf (nfile, nautilus_get_icon_size_for_zoom_level (NAUTILUS_ZOOM_LEVEL_STANDARD), TRUE, 0);
    }
    else
    {
        if (!(*pb = gdk_pixbuf_new_from_file (thumb_loc, NULL)))
            g_warning ("could not load %s (%s)\n", *name, *description);
    }

    //nautilus_file_unref(nfile);
    g_object_unref (fac);
    g_free (thumb_loc);
    g_free (uri);
}

void 
clutter_cover_flow_set_info_callback(ClutterCoverFlow *self)
{
    ClutterCoverFlowGetInfoCallback cb;

    g_return_if_fail( CLUTTER_IS_COVER_FLOW(self) );
    #if USE_LIBGNOME_FOR_THUMB
	    cb = get_info_nautilus_thumb;
    	cb = get_info_libgnome_thumb;
    #else
	    cb = get_info_libgnome_thumb;
    	cb = get_info_nautilus_thumb;
    #endif

    self->priv->get_info_callback = cb;
}

ClutterCoverFlow*
clutter_cover_flow_new_with_model (ClutterActor *stage, GtkTreeModel *store, GtkTreeView *tree)
{
    ClutterCoverFlow *self;
    ClutterColor color = { 255, 255, 255, 255 }; /* white */

    g_return_val_if_fail(CLUTTER_IS_STAGE(stage), NULL);
    g_return_val_if_fail(GTK_IS_TREE_MODEL(store), NULL);

    self = g_object_new (CLUTTER_TYPE_COVER_FLOW, NULL);
    clutter_cover_flow_set_model(self, tree, store);

    /* Add ourselves to the stage */
    self->priv->m_stage = stage;
    clutter_container_add_actor ( CLUTTER_CONTAINER (stage), CLUTTER_ACTOR(self) );

    /* Add a container, that will hold all covers, as our child */
    self->priv->m_container = clutter_group_new();
    clutter_container_add_actor ( CLUTTER_CONTAINER (self), self->priv->m_container );

    /* Add some text as our child */
    self->priv->item_name = clutter_text_new_full ("Lucida Grande bold 13", "", &color);
    clutter_container_add_actor (CLUTTER_CONTAINER (self->priv->m_container), self->priv->item_name);
    self->priv->item_type = clutter_text_new_full ("Lucida Grande 10", "", &color);
    clutter_container_add_actor (CLUTTER_CONTAINER (self->priv->m_container), self->priv->item_type);

    /* Track stage resizes. */
    g_signal_connect (
                      stage,
                      "notify::allocation",
                      G_CALLBACK (on_stage_resized),
                      self);
    clutter_cover_flow_set_info_callback(self);

    /* Fake resize event to set item initial position */
    on_stage_resized(CLUTTER_STAGE(stage), NULL, self);

    return self;
}

/* FIXME: new_with_model do a common fct (init or something */
ClutterCoverFlow*
clutter_cover_flow_new_with_model_icon (ClutterActor *stage, NautilusIconContainer *container)
{
    ClutterCoverFlow *self;
    ClutterColor color = { 255, 255, 255, 255 }; /* white */

    g_return_val_if_fail(CLUTTER_IS_STAGE(stage), NULL);
    g_return_val_if_fail(NAUTILUS_IS_ICON_CONTAINER(container), NULL);

    self = g_object_new (CLUTTER_TYPE_COVER_FLOW, NULL);
    clutter_cover_flow_set_model_icon(self, container);

    /* Add ourselves to the stage */
    self->priv->m_stage = stage;
    clutter_container_add_actor ( CLUTTER_CONTAINER (stage), CLUTTER_ACTOR(self) );

    /* Add a container, that will hold all covers, as our child */
    self->priv->m_container = clutter_group_new();
    clutter_container_add_actor ( CLUTTER_CONTAINER (self), self->priv->m_container );

    /* Add some text as our child */
    self->priv->item_name = clutter_text_new_full ("Lucida Grande bold 13", "", &color);
    clutter_container_add_actor (CLUTTER_CONTAINER (self->priv->m_container), self->priv->item_name);
    self->priv->item_type = clutter_text_new_full ("Lucida Grande 10", "", &color);
    clutter_container_add_actor (CLUTTER_CONTAINER (self->priv->m_container), self->priv->item_type);

    /* Track stage resizes. */
    g_signal_connect (
                      stage,
                      "notify::allocation",
                      G_CALLBACK (on_stage_resized),
                      self);
    clutter_cover_flow_set_info_callback(self);

    /* Fake resize event to set item initial position */
    on_stage_resized(CLUTTER_STAGE(stage), NULL, self);

    //self->priv->test = "test";

    return self;
}

//amtest
#if 0
GtkWidget *
clutter_widget_init(void)
{
    ClutterActor    *stage;
    GtkWidget       *clutter;

	ClutterColor     stage_color = { 0x00, 0x00, 0x00, 0xff }; /* black */
    if (gtk_clutter_init (NULL, NULL) != CLUTTER_INIT_SUCCESS) {
        g_error ("Unable to initialize GtkClutter");
    	return NULL;
	}

    clutter = gtk_clutter_embed_new ();
    stage = gtk_clutter_embed_get_stage (GTK_CLUTTER_EMBED (clutter));
    clutter_stage_set_color (CLUTTER_STAGE (stage), &stage_color);
    return (clutter);
}
#endif

ClutterCoverFlow*
clutter_cover_flow_new (ClutterActor *stage)
{
    ClutterCoverFlow* cf;

    cf = clutter_cover_flow_new_with_model(
                                           stage,
                                           GTK_TREE_MODEL( gtk_list_store_new (1, G_TYPE_FILE) ),
                                           NULL);
    cf->priv->model_is_list_store = TRUE;
    return cf;
}

gboolean
clutter_cover_flow_isplaying(ClutterCoverFlow *coverflow)
{
    ClutterCoverFlowPrivate *priv;

    priv = coverflow->priv;
    if (clutter_timeline_is_playing(priv->m_timeline))
        return TRUE;
    return FALSE;
}

static void
clutter_cover_flow_move(ClutterCoverFlow *coverflow, move_t dir)
{
    ClutterCoverFlowPrivate *priv;

    g_return_if_fail( CLUTTER_IS_COVER_FLOW(coverflow) );

    priv = coverflow->priv;
    if ( ! model_is_empty(priv) ) {
        stop(coverflow);
        clear_behaviours(coverflow);
        view_move(priv, dir, TRUE);
    }
}

void clutter_cover_flow_left(ClutterCoverFlow *coverflow)
{
    //printf("MOVE_LEFT\n");
    // coverflow->priv->idx_visible_front++;
    clutter_cover_flow_move(coverflow, MOVE_LEFT);
}

void clutter_cover_flow_right(ClutterCoverFlow *coverflow)
{
    //printf("MOVE_RIGHT\n");
    // coverflow->priv->idx_visible_front--;
    clutter_cover_flow_move(coverflow, MOVE_RIGHT);
}

ClutterActor* clutter_cover_flow_get_actor_at_pos(ClutterCoverFlow *coverflow, guint x, guint y)
{
    ClutterCoverFlowPrivate *priv = coverflow->priv;
    return clutter_stage_get_actor_at_pos(CLUTTER_STAGE(priv->m_stage),CLUTTER_PICK_ALL,x,y);
}


void clutter_cover_flow_scroll_to_actor(ClutterCoverFlow *coverflow, ClutterActor *actor)
{
    //    GSequenceIter *iter;
    int pos;
    ClutterCoverFlowPrivate *priv;
    GtkTreePath *path;
    CoverFlowItem *item;
    GList *list;
    GList *selection;

    g_return_if_fail( CLUTTER_IS_COVER_FLOW(coverflow) );
    g_return_if_fail( CLUTTER_IS_ACTOR(actor) );

    priv = coverflow->priv;

    //g_critical("TODO: %s", G_STRFUNC);
    //clutter_cover_flow_scroll_to_position(coverflow, priv->idx_visible_front + pos);

    if ((pos = get_actor_pos(priv, actor)))
    {
        if (priv->model)
        {
            path = gtk_tree_path_new_from_indices(priv->idx_visible_front+pos, -1);
           	gtk_tree_view_set_cursor(priv->tree, path, NULL, FALSE);
        }
        else
            if (priv->container)
            {
                //item = item_grab_index(priv, priv->idx_visible_front+pos);
                //printf("!!%s: %s\n", G_STRFUNC, nautilus_file_get_uri(item->file));
                list = g_list_nth(priv->icons, priv->idx_visible_front+pos);
                if (list != NULL)
                {
                    //char *uri = nautilus_icon_container_get_icon_uri(priv->container, list->data);
                    //printf("!!%s: %s\n", G_STRFUNC, uri);
                    selection = NULL;
                    NautilusIcon *icon;

                    icon = list->data;
                    selection = g_list_prepend (selection, icon->data);
                    /*printf("!!%s: %s\n", G_STRFUNC,
                           nautilus_file_get_uri(selection->data));*/
                    nautilus_icon_container_set_selection(priv->container, selection);
                }
            }
    }
}

void 
clutter_cover_flow_scroll_to_position(ClutterCoverFlow *cf, int pos)
{

    ClutterCoverFlowPrivate *priv;
    priv = cf->priv;


    int distance_to_move = pos - priv->idx_visible_front;
    //g_message("From %i to %i (%i)",priv->idx_visible_front, pos, distance_to_move);

    int j;

    if(distance_to_move>=0) /*Move left <--- */
    {
        for(j=0; j< distance_to_move; j++)
            clutter_cover_flow_left(cf);
    }
    else                    /*Move Right --->*/
    {
        distance_to_move *= -1;
        for(j=0; j< distance_to_move; j++)
            clutter_cover_flow_right(cf);
    }

}

void clutter_cover_flow_clear(ClutterCoverFlow *coverflow)
{
    ClutterCoverFlowPrivate *priv;

    g_return_if_fail( CLUTTER_IS_COVER_FLOW(coverflow) );

    stop(coverflow);
    clear_behaviours(coverflow);

    priv = coverflow->priv;
    if ( ! model_is_empty(priv) )
        knock_down_items(priv, TRUE);

}

void clutter_cover_flow_select(ClutterCoverFlow *coverflow, gboolean should_clear)
{
    ClutterCoverFlowPrivate *priv;

    g_return_if_fail( CLUTTER_IS_COVER_FLOW(coverflow) );

    stop(coverflow);
    clear_behaviours(coverflow);

    priv = coverflow->priv;
    if ( ! model_is_empty(priv) )
        zoom_items(priv, 2.0, should_clear);
}

void clutter_cover_flow_default_view(ClutterCoverFlow *coverflow)
{
    ;
}

void clutter_cover_flow_grid_view(ClutterCoverFlow *coverflow)
{
    ;
}

void clutter_cover_flow_film_view(ClutterCoverFlow *coverflow)
{
    ;
}

int clutter_cover_flow_get_front_index(ClutterCoverFlow *coverflow)
{
    return coverflow->priv->idx_visible_front;
}

